#pragma once

#ifdef __cplusplus
extern "C" {
#endif
  
  double poly_sin(double x);
  double poly_cos(double x);
  double poly_tan(double x);
  double poly_atan2(double y, double x);
  double poly_asin(double x);
  double poly_acos(double x);
  
#ifdef __cplusplus
}
#endif
